@@ -9,7 +9,7 @@
 #include <strings.h>
 #include <wayland-client.h>
 #include "background-image.h"
-#include "cairo.h"
+#include "cairo_util.h"
 #include "log.h"
 #include "pool-buffer.h"
 #include "wlr-layer-shell-unstable-v1-client-protocol.h"
@@ -41,12 +41,20 @@ struct swaybg_state {
 	struct zxdg_output_manager_v1 *xdg_output_manager;
 	struct wl_list configs;  // struct swaybg_output_config::link
 	struct wl_list outputs;  // struct swaybg_output::link
+	struct wl_list images;   // struct swaybg_image::link
 	bool run_display;
 };
 
+struct swaybg_image {
+	struct wl_list link;
+	const char *path;
+	bool load_required;
+};
+
 struct swaybg_output_config {
 	char *output;
-	cairo_surface_t *image;
+	const char *image_path;
+	struct swaybg_image *image;
 	enum background_mode mode;
 	uint32_t color;
 	struct wl_list link;
@@ -64,12 +72,14 @@ struct swaybg_output {
 
 	struct wl_surface *surface;
 	struct zwlr_layer_surface_v1 *layer_surface;
-	struct pool_buffer buffers[2];
-	struct pool_buffer *current_buffer;
 
 	uint32_t width, height;
 	int32_t scale;
 
+	uint32_t configure_serial;
+	bool dirty, needs_ack;
+	int32_t committed_width, committed_height, committed_scale;
+
 	struct wl_list link;
 };
 
@@ -91,15 +101,29 @@ bool is_valid_color(const char *color) {
 	return true;
 }
 
-static void render_frame(struct swaybg_output *output) {
+static void render_frame(struct swaybg_output *output, cairo_surface_t *surface) {
 	int buffer_width = output->width * output->scale,
 		buffer_height = output->height * output->scale;
-	output->current_buffer = get_next_buffer(output->state->shm,
-			output->buffers, buffer_width, buffer_height);
-	if (!output->current_buffer) {
+
+	// If the last committed buffer has the same size as this one would, do
+	// not render a new buffer, because it will be identical to the old one
+	if (output->committed_width == buffer_width &&
+			output->committed_height == buffer_height) {
+		if (output->committed_scale != output->scale) {
+			wl_surface_set_buffer_scale(output->surface, output->scale);
+			wl_surface_commit(output->surface);
+
+			output->committed_scale = output->scale;
+		}
 		return;
 	}
-	cairo_t *cairo = output->current_buffer->cairo;
+	struct pool_buffer buffer;
+	if (!create_buffer(&buffer, output->state->shm,
+			buffer_width, buffer_height, WL_SHM_FORMAT_ARGB8888)) {
+		return;
+	}
+
+	cairo_t *cairo = buffer.cairo;
 	cairo_save(cairo);
 	cairo_set_operator(cairo, CAIRO_OPERATOR_CLEAR);
 	cairo_paint(cairo);
@@ -112,14 +136,32 @@ static void render_frame(struct swaybg_output *output) {
 			cairo_set_source_u32(cairo, output->config->color);
 			cairo_paint(cairo);
 		}
-		render_background_image(cairo, output->config->image,
+
+		if (surface) {
+			render_background_image(cairo, surface,
 				output->config->mode, buffer_width, buffer_height);
+		}
 	}
 
 	wl_surface_set_buffer_scale(output->surface, output->scale);
-	wl_surface_attach(output->surface, output->current_buffer->buffer, 0, 0);
+	wl_surface_attach(output->surface, buffer.buffer, 0, 0);
 	wl_surface_damage_buffer(output->surface, 0, 0, INT32_MAX, INT32_MAX);
 	wl_surface_commit(output->surface);
+
+	output->committed_width = buffer_width;
+	output->committed_height = buffer_height;
+	output->committed_scale = output->scale;
+
+	// we will not reuse the buffer, so destroy it immediately
+	destroy_buffer(&buffer);
+}
+
+static void destroy_swaybg_image(struct swaybg_image *image) {
+	if (!image) {
+		return;
+	}
+	wl_list_remove(&image->link);
+	free(image);
 }
 
 static void destroy_swaybg_output_config(struct swaybg_output_config *config) {
@@ -144,8 +186,6 @@ static void destroy_swaybg_output(struct swaybg_output *output) {
 	}
 	zxdg_output_v1_destroy(output->xdg_output);
 	wl_output_destroy(output->wl_output);
-	destroy_buffer(&output->buffers[0]);
-	destroy_buffer(&output->buffers[1]);
 	free(output->name);
 	free(output->identifier);
 	free(output);
@@ -157,8 +197,9 @@ static void layer_surface_configure(void *data,
 	struct swaybg_output *output = data;
 	output->width = width;
 	output->height = height;
-	zwlr_layer_surface_v1_ack_configure(surface, serial);
-	render_frame(output);
+	output->dirty = true;
+	output->configure_serial = serial;
+	output->needs_ack = true;
 }
 
 static void layer_surface_closed(void *data,
@@ -194,7 +235,7 @@ static void output_scale(void *data, struct wl_output *wl_output,
 	struct swaybg_output *output = data;
 	output->scale = scale;
 	if (output->state->run_display && output->width > 0 && output->height > 0) {
-		render_frame(output);
+		output->dirty = true;
 	}
 }
 
@@ -367,10 +408,8 @@ static bool store_swaybg_output_config(struct swaybg_state *state,
 	wl_list_for_each(oc, &state->configs, link) {
 		if (strcmp(config->output, oc->output) == 0) {
 			// Merge on top
-			if (config->image) {
-				free(oc->image);
-				oc->image = config->image;
-				config->image = NULL;
+			if (config->image_path) {
+				oc->image_path = config->image_path;
 			}
 			if (config->color) {
 				oc->color = config->color;
@@ -432,11 +471,7 @@ static void parse_command_line(int argc, char **argv,
 			config->color = parse_color(optarg);
 			break;
 		case 'i':  // image
-			free(config->image);
-			config->image = load_background_image(optarg);
-			if (!config->image) {
-				swaybg_log(LOG_ERROR, "Failed to load image: %s", optarg);
-			}
+			config->image_path = optarg;
 			break;
 		case 'm':  // mode
 			config->mode = parse_background_mode(optarg);
@@ -486,10 +521,10 @@ static void parse_command_line(int argc, char **argv,
 	config = NULL;
 	struct swaybg_output_config *tmp = NULL;
 	wl_list_for_each_safe(config, tmp, &state->configs, link) {
-		if (!config->image && !config->color) {
+		if (!config->image_path && !config->color) {
 			destroy_swaybg_output_config(config);
 		} else if (config->mode == BACKGROUND_MODE_INVALID) {
-			config->mode = config->image
+			config->mode = config->image_path
 				? BACKGROUND_MODE_STRETCH
 				: BACKGROUND_MODE_SOLID_COLOR;
 		}
@@ -502,9 +537,32 @@ int main(int argc, char **argv) {
 	struct swaybg_state state = {0};
 	wl_list_init(&state.configs);
 	wl_list_init(&state.outputs);
+	wl_list_init(&state.images);
 
 	parse_command_line(argc, argv, &state);
 
+	// Identify distinct image paths which will need to be loaded
+	struct swaybg_image *image;
+	struct swaybg_output_config *config;
+	wl_list_for_each(config, &state.configs, link) {
+		if (!config->image_path) {
+			continue;
+		}
+		wl_list_for_each(image, &state.images, link) {
+			if (strcmp(image->path, config->image_path) == 0) {
+				config->image = image;
+				break;
+			}
+		}
+		if (config->image) {
+			continue;
+		}
+		image = calloc(1, sizeof(struct swaybg_image));
+		image->path = config->image_path;
+		wl_list_insert(&state.images, &image->link);
+		config->image = image;
+	}
+
 	state.display = wl_display_connect(NULL);
 	if (!state.display) {
 		swaybg_log(LOG_ERROR, "Unable to connect to the compositor. "
@@ -532,7 +590,55 @@ int main(int argc, char **argv) {
 
 	state.run_display = true;
 	while (wl_display_dispatch(state.display) != -1 && state.run_display) {
-		// This space intentionally left blank
+		// Send acks, and determine which images need to be loaded
+		wl_list_for_each(output, &state.outputs, link) {
+			if (output->needs_ack) {
+				output->needs_ack = false;
+				zwlr_layer_surface_v1_ack_configure(
+						output->layer_surface,
+						output->configure_serial);
+			}
+
+			int buffer_width = output->width * output->scale,
+				buffer_height = output->height * output->scale;
+			bool buffer_change =
+				output->committed_height != buffer_height ||
+				output->committed_width != buffer_width;
+			if (output->dirty && output->config->image && buffer_change) {
+				output->config->image->load_required = true;
+			}
+		}
+
+		// Load images, render associated frames, and unload
+		wl_list_for_each(image, &state.images, link) {
+			if (!image->load_required) {
+				continue;
+			}
+
+			cairo_surface_t *surface = load_background_image(image->path);
+			if (!surface) {
+				swaybg_log(LOG_ERROR, "Failed to load image: %s", image->path);
+				continue;
+			}
+
+			wl_list_for_each(output, &state.outputs, link) {
+				if (output->dirty && output->config->image == image) {
+					output->dirty = false;
+					render_frame(output, surface);
+				}
+			}
+
+			image->load_required = false;
+			cairo_surface_destroy(surface);
+		}
+
+		// Redraw outputs without associated image
+		wl_list_for_each(output, &state.outputs, link) {
+			if (output->dirty) {
+				output->dirty = false;
+				render_frame(output, NULL);
+			}
+		}
 	}
 
 	struct swaybg_output *tmp_output;
@@ -540,10 +646,15 @@ int main(int argc, char **argv) {
 		destroy_swaybg_output(output);
 	}
 
-	struct swaybg_output_config *config = NULL, *tmp_config = NULL;
+	struct swaybg_output_config *tmp_config = NULL;
 	wl_list_for_each_safe(config, tmp_config, &state.configs, link) {
 		destroy_swaybg_output_config(config);
 	}
 
+	struct swaybg_image *tmp_image;
+	wl_list_for_each_safe(image, tmp_image, &state.images, link) {
+		destroy_swaybg_image(image);
+	}
+
 	return 0;
 }
